# -------------------------------------------------------------
# NDN Hydra Add Group Message
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Date:    2021-01-25
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

from typing import Callable
from ndn.encoding import *
import time
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.group_messages.specific_message import SpecificMessage
from ndn_hydra.repo.protocol.base_models import File

class AddMessageTypes:
    NODE_NAME = 84
    FAVOR = 86

    FILE = 91
    DESIRED_COPIES = 92
    FETCH_PATH = 93
    IS_STORED_BY_ORIGIN = 94

    BACKUP = 100
    BACKUP_NODE_NAME = 101
    BACKUP_NONCE = 102

class FetchPathTlv(TlvModel):
    prefix = NameField()

class BackupTlv(TlvModel):
    node_name = BytesField(AddMessageTypes.BACKUP_NODE_NAME)
    nonce = BytesField(AddMessageTypes.BACKUP_NONCE)

class AddMessageTlv(TlvModel):
    node_name = BytesField(AddMessageTypes.NODE_NAME)
    favor = BytesField(AddMessageTypes.FAVOR)
    file = ModelField(AddMessageTypes.FILE, File)

    desired_copies = UintField(AddMessageTypes.DESIRED_COPIES)
    fetch_path = ModelField(AddMessageTypes.FETCH_PATH, FetchPathTlv)
    is_stored_by_origin = UintField(AddMessageTypes.IS_STORED_BY_ORIGIN)
    backup_list = RepeatedField(ModelField(AddMessageTypes.BACKUP, BackupTlv))

class AddMessage(SpecificMessage):
    def __init__(self, nid:str, seqno:int, raw_bytes:bytes):
        super(AddMessage, self).__init__(nid, seqno)
        self.message = AddMessageTlv.parse(raw_bytes)

    async def apply(self, global_view: GlobalView, fetch_file: Callable, svs, config):
        node_name = self.message.node_name.tobytes().decode()
        favor = float(self.message.favor.tobytes().decode())
        file = self.message.file
        file_name = Name.to_str(file.file_name)
        packets = file.packets
        digests = file.digests
        size = file.size
        desired_copies = self.message.desired_copies
        fetch_path = self.message.fetch_path.prefix
        is_stored_by_origin = False if (self.message.is_stored_by_origin == 0) else True
        backups = self.message.backup_list
        backup_list = []
        bak = ""
        for backup in backups:
            backup_list.append((backup.node_name.tobytes().decode(), backup.nonce.tobytes().decode()))
            bak = bak + backup.node_name.tobytes().decode() + ","
        self.logger.info(f"[MSG][ADD]      nam={node_name};fil={file_name};cop={desired_copies};pck={packets};siz={size};bak={bak}")
        global_view.add_file(
            file_name,
            size,
            node_name,
            Name.to_str(fetch_path),
            self.seqno,
            b''.join(digests),
            packets=packets,
            desired_copies=desired_copies
        )

        global_view.set_backups(file_name, backup_list)

        # get pending stores
        copies_needed = desired_copies
        pending_stores = global_view.get_pending_stores(file_name)
        for pending_store in pending_stores:
            # data_storage.add_metainfos(insertion_id, Name.to_str(file_name), packets, digests, Name.to_str(fetch_path))
            global_view.store_file(file_name, pending_store)
            copies_needed -= 1

        # if I need to store this file
        # if is_stored_by_origin:
        #     copies_needed -= 1
        need_to_store = False
        for i in range(copies_needed):
            backup = backup_list[i]
            if backup[0] == config['node_name']:
                need_to_store = True
                break
        if need_to_store == True:
            fetch_file(file_name, packets, digests, Name.to_str(fetch_path))

            # from .message import MessageTlv, MessageTypes
            # # generate store msg and send
            # # store tlv
            # expire_at = int(time.time()+(config['period']*2))
            # favor = 1.85
            # store_message = StoreMessageTlv()
            # store_message.session_id = config['session_id'].encode()
            # store_message.node_name = config['node_name'].encode()
            # store_message.expire_at = expire_at
            # store_message.favor = str(favor).encode()
            # store_message.insertion_id = insertion_id.encode()
            # # store msg
            # store_message = MessageTlv()
            # store_message.type = MessageTypes.STORE
            # store_message.value = store_message.encode()
            # # apply globalview and send msg thru SVS
            # # next_state_vector = svs.getCore().getStateVector().get(config['session_id']) + 1

            # # global_view.store_file(insertion_id, config['session_id'])
            # svs.publishData(store_message.encode())
            # val = "[MSG][STORE]*  sid={sid};iid={iid}".format(
            #     sid=config['session_id'],
            #     iid=insertion_id
            # )
            # self.logger.info(val)
        # update session
        global_view.update_node(node_name, favor, self.seqno)


# -------------------------------------------------------------
# NDN Hydra Query Client
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Date:    2021-01-25
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import logging
from ndn.app import NDNApp
from ndn.encoding import FormalName, Component, Name, ContentType
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
from ndn_hydra.repo.protocol.base_models import FileList, File

class HydraQueryClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
      """
      This client queries a node within the remote repo.
      :param app: NDNApp.
      :param client_prefix: NonStrictName. Routable name to client.
      :param repo_prefix: NonStrictName. Routable name to remote repo.
      """
      self.app = app
      self.client_prefix = client_prefix
      self.repo_prefix = repo_prefix

    async def send_query(self, query: Name, node_name: str=None) -> None:
      """
      Form a certain query and request that info from a node.
      """
      if not node_name:
          named_query = self.repo_prefix + [Component.from_str("query")] + query
      else:
          named_query = self.repo_prefix + [Component.from_str("node")] + [Component.from_str(node_name)] + [Component.from_str("query")] + query

      try:
          data_name, meta_info, content = await self.app.express_interest(named_query, can_be_prefix=True, must_be_fresh=True, lifetime=3000)
          if meta_info.content_type == ContentType.NACK:
             print("Distributed Repo does not know that query.")
             return
          else:
             querytype = Component.to_str(query[0])
             if querytype == "nodes":
                 print(f'List of All Node Names')
                 print(f'{bytes(content).decode().split()}')
                 return
             elif querytype == "exnodes":
                 print(f'List of All Expired Node Names')
                 print(f'{bytes(content).decode().split()}')
                 return
             elif querytype == "files":
                 filelist = FileList.parse(content)
                 counter = 1
                 if filelist.list:
                     print(f'List of All Files')
                     for file in filelist.list:
                         print(f'File {counter} meta-info')
                         print(f'\tfile_name: {Name.to_str(file.file_name)}')
                         print(f'\tsize: {file.size}')
                         counter = counter + 1
                 else:
                     print(f'No files inserted in the remote repo.')
                 return
             elif querytype == "file":
                 if content:
                     file = File.parse(content)
                     print(f'File Exists, File meta-info')
                     print(f'\tfile_name: {Name.to_str(file.file_name)}')
                     print(f'\tsize: {file.size}')
                 else:
                     print(f'File Does Not Exists in The Repo')
                 return
             elif querytype == "prefix":
                 filelist = FileList.parse(content)
                 counter = 1
                 if filelist.list:
                     print(f'List of All Files with prefix {Name.to_str(query[1:])}')
                     for file in filelist.list:
                         print(f'File {counter} meta-info')
                         print(f'\tfile_name: {Name.to_str(file.file_name)}')
                         print(f'\tsize: {file.size}')
                         counter = counter + 1
                 else:
                     print(f'No files inserted in the remote repo with prefix {Name.to_str(query[1:])}.')
                 return
             else:
                 print("Client does not know that query.")
                 return
      except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure) as e:
          print("Query command received no data packet back")
          return



# -------------------------------------------------------------
# NDN Hydra Delete Client
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Date:    2021-01-25
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import logging
from ndn.app import NDNApp
from ndn.encoding import Name, Component, FormalName
from ndn_hydra.repo.protocol.base_models import DeleteCommand
from ndn_hydra.repo.utils.pubsub import PubSub

class HydraDeleteClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
      """
      This client deletes a certain file from the remote repo.
      :param app: NDNApp.
      :param client_prefix: NonStrictName. Routable name to client.
      :param repo_prefix: NonStrictName. Routable name to remote repo.
      """
      self.app = app
      self.client_prefix = client_prefix
      self.repo_prefix = repo_prefix
      self.pb = PubSub(self.app, self.client_prefix)

    async def delete_file(self, file_name: FormalName) -> bool:
      """
      Delete a file asscoiated with a file name from the remote repo
      """
      # send command interest
      cmd = DeleteCommand()
      cmd.file_name = file_name
      cmd_bytes = cmd.encode()

      # publish msg to repo's delete topic
      await self.pb.wait_for_ready()
      is_success = await self.pb.publish(self.repo_prefix + ['delete'], cmd_bytes)
      if is_success:
          logging.debug('Published an delete msg and was acknowledged by a subscriber')
      else:
          logging.debug('Published an delete msg but was not acknowledged by a subscriber')
      return is_success



# -------------------------------------------------------------
# NDN Hydra Claim Group Message
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Date:    2021-01-25
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

import copy
from typing import Callable
import time
from ndn.encoding import *
from ndn_hydra.repo.modules.global_view import GlobalView
from ndn_hydra.repo.group_messages.specific_message import SpecificMessage

class ClaimTypes:
    REQUEST = 1
    COMMITMENT = 2
class ClaimMessageTypes:
    NODE_NAME = 84
    FAVOR = 86

    TYPE = 91 # 1=request; 2=commitment
    CLAIMER_NODE_NAME = 92
    CLAIMER_NONCE = 93
    AUTHORIZER_NODE_NAME = 94
    AUTHORIZER_NONCE = 95


class ClaimMessageTlv(TlvModel):
    node_name = BytesField(ClaimMessageTypes.NODE_NAME)
    favor = BytesField(ClaimMessageTypes.FAVOR)
    file_name = NameField()
    type = UintField(ClaimMessageTypes.TYPE)
    claimer_node_name = BytesField(ClaimMessageTypes.CLAIMER_NODE_NAME)
    claimer_nonce = BytesField(ClaimMessageTypes.CLAIMER_NONCE)
    authorizer_node_name = BytesField(ClaimMessageTypes.AUTHORIZER_NODE_NAME)
    authorizer_nonce = BytesField(ClaimMessageTypes.AUTHORIZER_NONCE)

class ClaimMessage(SpecificMessage):
    def __init__(self, nid:str, seqno:int, raw_bytes:bytes):
        super(ClaimMessage, self).__init__(nid, seqno)
        self.message = ClaimMessageTlv.parse(raw_bytes)

    async def apply(self, global_view, fetch_file, svs, config):
        node_name = self.message.node_name.tobytes().decode()
        favor = float(self.message.favor.tobytes().decode())
        file_name = Name.to_str(self.message.file_name)
        claimer_node_name = self.message.claimer_node_name.tobytes().decode()
        claimer_nonce = self.message.claimer_nonce.tobytes().decode()
        authorizer_node_name = self.message.authorizer_node_name.tobytes().decode()
        authorizer_nonce = self.message.authorizer_nonce.tobytes().decode()
        file = global_view.get_file(file_name)
        if self.message.type == ClaimTypes.COMMITMENT:
            rank = len(file['backuped_bys'])
            self.logger.info(f"[MSG][CLAIM.C]  nam={claimer_node_name};fil={file_name}")
            global_view.add_backup(file_name, claimer_node_name, rank, claimer_nonce)
        else:
            self.logger.info(f"[MSG][CLAIM.R]  nam={claimer_node_name};fil={file_name}")
            if authorizer_node_name == config['node_name']:
                from .message import MessageTlv, MessageTypes
                commit = False
                if (len(file['backuped_bys']) == 0) and (file['stored_bys'][-1] == config['node_name']) and (authorizer_nonce == file['id']):
                    global_view.add_backup(file_name, claimer_node_name, 0, claimer_nonce)
                    commit = True
                if (len(file['backuped_bys']) > 0) and (file['backuped_bys'][-1]['node_name'] == config['node_name']) and (authorizer_nonce == file['backuped_bys'][-1]['nonce']):
                    global_view.add_backup(file_name, claimer_node_name, len(file['backuped_bys']), claimer_nonce)
                    commit = True
                if commit == True:
                    # claim tlv
                    favor = 1.85
                    claim_message = copy.copy(self.message)
                    claim_message.node_name = config['node_name'].encode()
                    claim_message.favor = str(favor).encode()
                    claim_message.type = ClaimTypes.COMMITMENT
                    # claim msg
                    claim_message = MessageTlv()
                    claim_message.type = MessageTypes.CLAIM
                    claim_message.value = claim_message.encode()
                    svs.publishData(claim_message.encode())
                    self.logger.info(f"[MSG][CLAIM.C]* nam={claimer_node_name};fil={file_name}")
        global_view.update_node(node_name, favor, self.seqno)


# -------------------------------------------------------------
# NDN Hydra General Group Message
# -------------------------------------------------------------
#  @Project: NDN Hydra
#  @Date:    2021-01-25
#  @Authors: Please check AUTHORS.rst
#  @Source-Code:   https://github.com/justincpresley/ndn-hydra
#  @Documentation: https://ndn-hydra.readthedocs.io
#  @Pip-Library:   https://pypi.org/project/ndn-hydra
# -------------------------------------------------------------

from __future__ import annotations

from ndn.encoding import *
from typing import Optional
from ndn_hydra.repo.protocol.tlv import HydraTlvTypes
from ndn_hydra.repo.group_messages.specific_message import SpecificMessage
from ndn_hydra.repo.group_messages.add import AddMessage
from ndn_hydra.repo.group_messages.remove import RemoveMessage
from ndn_hydra.repo.group_messages.store import StoreMessage
from ndn_hydra.repo.group_messages.claim import ClaimMessage
from ndn_hydra.repo.group_messages.heartbeat import HeartbeatMessage

class MessageTypes:
    ADD = 1
    REMOVE = 2
    STORE = 3
    CLAIM = 4
    HEARTBEAT = 5

class Message(TlvModel):
    type = UintField(HydraTlvTypes.MESSAGE_TYPE)
    value = BytesField(HydraTlvTypes.MESSAGE)
    @staticmethod
    def specify(nid:str, seqno:int, message_bytes:bytes) -> Optional[SpecificMessage]:
        message = Message.parse(message_bytes)
        message_type, message_bytes = message.type, bytes(message.value)
        if message_type == MessageTypes.ADD:
            return AddMessage(nid, seqno, message_bytes)
        elif message_type == MessageTypes.REMOVE:
            return RemoveMessage(nid, seqno, message_bytes)
        elif message_type == MessageTypes.STORE:
            return StoreMessage(nid, seqno, message_bytes)
        elif message_type == MessageTypes.CLAIM:
            return ClaimMessage(nid, seqno, message_bytes)
        elif message_type == MessageTypes.HEARTBEAT:
            return HeartbeatMessage(nid, seqno, message_bytes)
        else:
            return None